/*
 * libdigidocpp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "crypto/TSL.h"

#include "Conf.h"
#include "log.h"
#include "crypto/Connect.h"
#include "util/DateTime.h"
#include "util/File.h"
#include "xml/ts_119612v010101.hxx"

#ifdef __APPLE__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wnull-conversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
#endif
#include <xsec/enc/XSECKeyInfoResolverDefault.hpp>
#include <xsec/enc/OpenSSL/OpenSSLCryptoX509.hpp>
#include <xsec/framework/XSECException.hpp>
#include <xsec/framework/XSECProvider.hpp>
#ifdef __APPLE__
#pragma GCC diagnostic pop
#endif

#include <fstream>

using namespace digidoc;
using namespace digidoc::tsl;
using namespace digidoc::util;
using namespace digidoc::util::date;
using namespace std;
using namespace xercesc;
using namespace xml_schema;

TSL::TSL(const string &file)
{
    if(!File::fileExists(file))
        return;

    static const string SCHEMES_URI_V1 = "http://uri.etsi.org/TrstSvc/eSigDir-1999-93-EC-TrustedList/TSLType/schemes";
    static const string SCHEMES_URI_V2 = "http://uri.etsi.org/TrstSvc/TrustedList/TSLType/EUlistofthelists";
    static const string GENERIC_URI_V1 = "http://uri.etsi.org/TrstSvc/eSigDir-1999-93-EC-TrustedList/TSLType/generic";
    static const string GENERIC_URI_V3 = "http://uri.etsi.org/TrstSvc/TSLtype/generic/eSigDir-1999-93-EC-TrustedList";
    static const string GENERIC_URI_V2 = "http://uri.etsi.org/TrstSvc/TrustedList/TSLType/EUgeneric";
    static const string CAQC_URI = "http://uri.etsi.org/TrstSvc/Svctype/CA/QC";
    static const string OCSP_URI = "http://uri.etsi.org/TrstSvc/Svctype/Certstatus/OCSP";
    static const string OCSPQC_URI = "http://uri.etsi.org/TrstSvc/Svctype/Certstatus/OCSP/QC";
    static const string TSA_URI = "http://uri.etsi.org/TrstSvc/Svctype/TSA";

    try {
        Properties properties;
        properties.schema_location("http://uri.etsi.org/02231/v2#", Conf::instance()->xsdPath() + "/ts_119612v010101.xsd");
        tsl = trustServiceStatusList(file,
            Flags::keep_dom|Flags::dont_initialize|Flags::dont_validate, properties);

        const TSLSchemeInformationType &info = tsl->schemeInformation();
        type = info.tSLType();
        operatorName = toString(info.schemeOperatorName());
        //BIO_printf(bio, "Name: %s\n", toString(info.schemeName()).c_str());
        territory = info.schemeTerritory().get();
        issueDate = xsd2string(info.listIssueDateTime());
        nextUpdate = xsd2string(info.nextUpdate().dateTime().get());

        if((info.tSLType() == SCHEMES_URI_V1 || info.tSLType() == SCHEMES_URI_V2) &&
                info.pointersToOtherTSL().present())
        {
            for(const OtherTSLPointersType::OtherTSLPointerType &other:
                info.pointersToOtherTSL()->otherTSLPointer())
            {
                if(!other.additionalInformation().present() ||
                   !other.serviceDigitalIdentities().present() ||
                   other.additionalInformation()->mimeType() != "application/vnd.etsi.tsl+xml")
                    continue;

                Pointer p;
                p.territory = other.additionalInformation()->schemeTerritory();
                p.location = string(other.tSLLocation());
                for(const ServiceDigitalIdentityListType::ServiceDigitalIdentityType &identity:
                    other.serviceDigitalIdentities()->serviceDigitalIdentity())
                {
                    for(const DigitalIdentityListType::DigitalIdType &id: identity.digitalId())
                    {
                        if(!id.x509Certificate().present())
                            continue;
                        Base64Binary base64 = id.x509Certificate().get();
                        p.certs.push_back(X509Cert(vector<unsigned char>(base64.data(), base64.data() + base64.capacity())));
                    }
                }
                pointer.push_back(p);
            }
        }
        else if((info.tSLType() == GENERIC_URI_V1 || info.tSLType() == GENERIC_URI_V2 || info.tSLType() == GENERIC_URI_V3) &&
                tsl->trustServiceProviderList().present())
        {
            for(const TrustServiceProviderListType::TrustServiceProviderType &pointer:
                tsl->trustServiceProviderList()->trustServiceProvider())
            {
                for(const TSPServicesListType::TSPServiceType &service:
                    pointer.tSPServices().tSPService())
                {
                    const TSPServiceInformationType &serviceInfo = service.serviceInformation();
                    if(serviceInfo.serviceTypeIdentifier() != CAQC_URI &&
                       serviceInfo.serviceTypeIdentifier() != OCSP_URI &&
                       serviceInfo.serviceTypeIdentifier() != OCSPQC_URI &&
                       serviceInfo.serviceTypeIdentifier() != TSA_URI)
                        continue;

                    //BIO_printf(bio, "Provider name: %sn\n", toString(i->tSPInformation().tSPName()).c_str());
                    //BIO_printf(bio, " Serivce type: %s\n", serviceInfo.serviceTypeIdentifier().c_str());
                    //BIO_printf(bio, " Service name: %s\n", toString(serviceInfo.serviceName()).c_str());
                    for(const DigitalIdentityListType::DigitalIdType id:
                        serviceInfo.serviceDigitalIdentity().digitalId())
                    {
                        if(!id.x509Certificate().present())
                            continue;
                        Base64Binary base64 = id.x509Certificate().get();
                        certs.push_back(X509Cert(vector<unsigned char>(base64.data(), base64.data() + base64.capacity())));
                    }

                    if(!service.serviceHistory().present())
                        continue;
                    for(const ServiceHistoryInstanceType &history: service.serviceHistory()->serviceHistoryInstance())
                    {
                        if(history.serviceTypeIdentifier() != CAQC_URI &&
                           history.serviceTypeIdentifier() != OCSP_URI &&
                           history.serviceTypeIdentifier() != OCSPQC_URI &&
                           history.serviceTypeIdentifier() != TSA_URI)
                            continue;

                        for(const DigitalIdentityListType::DigitalIdType id:
                            history.serviceDigitalIdentity().digitalId())
                        {
                            if(!id.x509Certificate().present())
                                continue;
                            Base64Binary base64 = id.x509Certificate().get();
                            certs.push_back(X509Cert(vector<unsigned char>(base64.data(), base64.data() + base64.capacity())));
                        }
                    }
                }
            }
        }
    }
    catch(const Parsing &e)
    {
        stringstream s;
        s << e;
        WARN("Failed to parse TSL %s %s: %s", territory.c_str(), file.c_str(), s.str().c_str());
    }
    catch(const xsd::cxx::exception &e)
    {
        WARN("Failed to parse TSL %s %s: %s", territory.c_str(), file.c_str(), e.what());
    }
    catch(XMLException &e)
    {
        string msg = xsd::cxx::xml::transcode<char>(e.getMessage());
        WARN("Failed to parse TSL %s %s: %s", territory.c_str(), file.c_str(), msg.c_str());
    }
    catch(const Exception &e)
    {
        WARN("Failed to parse TSL %s %s: %s", territory.c_str(), file.c_str(), e.msg().c_str());
    }
    catch(...)
    {
        WARN("Failed to parse TSL %s %s", territory.c_str(), file.c_str());
    }
}

TSL::~TSL()
{
}

void TSL::parse(vector<X509Cert> &list)
{
    ConfV2 *c = ConfV2::instance();
    string cache = c ? c->TSLCache() : ConfV2().TSLCache();
    string url = c ? c->TSLUrl() : ConfV2().TSLUrl();
    std::vector<X509Cert> cert;
    cert.push_back(c ? c->TSLCert() : ConfV2().TSLCert());
    File::createDirectory(cache);
    parse(list, url, cert, cache, File::fileName(url));
}

void TSL::parse(vector<X509Cert> &list, const string &url, const vector<X509Cert> &certs, const string &cache, const string &territory)
{
    string path = cache + territory;
    TSL tsl(path);

    time_t t = time(0);
    struct tm *time = gmtime(&t);
    if(tsl.nextUpdate.empty() ||
       tsl.nextUpdate.compare(13, 20, xsd2string(makeDateTime(*time))) <= 0)
    {
        Connect c(url, "GET");
        ofstream file(File::encodeName(path).c_str(), ofstream::binary);
        file << c.read();
        file.close();
        tsl = TSL(path);
    }

    bool found = false;
    for(const X509Cert &cert: certs)
    {
        try {
            tsl.validate(cert);
            found = true;
        } catch(...) {}
    }
    if(!found)
        WARN("TSL %s signature is invalid", tsl.territory.c_str());


    if(!tsl.pointer.empty())
    {
        for(const TSL::Pointer &p: tsl.pointer)
        {
            if(p.territory == "EE" || p.territory == "FI" || p.territory == "LV" || p.territory == "LT")
                parse(list, p.location, p.certs, cache, p.territory + ".xml");
        }
    }
    else
        list.insert(list.end(), tsl.certs.begin(), tsl.certs.end());
}

string TSL::toString(const InternationalNamesType &obj, const string &lang) const
{
    for(const InternationalNamesType::NameType &name: obj.name())
        if(name.lang() == lang)
            return name;
    return obj.name().front();
}

void TSL::validate(const X509Cert &cert)
{
    if(!tsl)
        THROW_CAUSE(Exception(EXCEPTION_PARAMS("Failed to parse XML")), "TSL Signature is invalid");

    try {
        XSECProvider prov;
        DSIGSignature *sig = prov.newSignatureFromDOM(tsl->_node()->getOwnerDocument());
        OpenSSLCryptoX509 x509(cert.handle());
        //sig->setKeyInfoResolver(new XSECKeyInfoResolverDefault);
        sig->setSigningKey(x509.clonePublicKey());
        //sig->registerIdAttributeName(MAKE_UNICODE_STRING("ID"));
        sig->load();
        if(!sig->verify())
        {
            string msg = xsd::cxx::xml::transcode<char>(sig->getErrMsgs());
            DEBUG("%s", msg.c_str());
            THROW("TLS Signature is invalid: %s", msg.c_str());
        }
    }
    catch(XSECException &e)
    {
        string msg = xsd::cxx::xml::transcode<char>(e.getMsg());
        DEBUG("%s", msg.c_str());
        THROW_CAUSE(Exception(EXCEPTION_PARAMS(msg.c_str())), "TSL Signature is invalid: %s", msg.c_str());
    }
    catch(const Exception &)
    {
        throw;
    }
    catch(...)
    {
        THROW("TSL Signature is invalid");
    }
}
